
lib_public {

    // vmat.l                                            view matrix support
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 06-14-96: wrote
    // 02-26-98: moved up to top-level dir from Terrain; changed _per
    // 10-21-99: moved into geo.l archive
    // 01-27-00: added vmat_write/_read (pack.l/.pk files),
    //           _evalpack interpreter
    // 09-09-00: moved win.l-dependent code to vmatwin.l
    // 12-17-00: added _scale()

    lib_include "../geo.l";             // for view stuff and frame_invert()

    typedef double (*vmatmat)[4];

    lib_struct {
        double m[4][4];
    } vmat;

    // usage:
    //
    // [x']   [m00 m01 m02 m03] [x]
    // [y'] = [m10 m11 m12 m13] [y]
    // [z']   [m20 m21 m22 m23] [z]
    // [w']   [m30 m31 m32 m33] [w]
    //
    // NOTE: this needs to be transposed into SGI view matrices

    vmat vmat_create();
    void vmat_destroy(vmat vm);
    void vmat_ident(vmat vmr);
    void vmat_scale(double q,vmat vmr);
    void vmat_move(double dx,double dy,double dz,vmat vmr);
    void vmat_xy(double a,vmat vmr);
    void vmat_yz(double a,vmat vmr);
    void vmat_zx(double a,vmat vmr);
    void vmat_per(double q,vmat vmr);
    void vmat_mul(vmat vma,vmat vmb,vmat vmr);
    void vmat_invert(vmat vm,vmat vmr);
    void vmat_copy(vmat vm,vmat vmr);
    void vmat_apply(vmat vm,double *p,double *pr);
    void vmat_print(vmat vm,str name);
    int  vmat_write(vmat vm,char *filename);
    vmat vmat_read(char *filename);
    vmat vmat_evalpack(pack pk);

};


vmat vmat_create()
{
    int i,j;
    vmat vm;

    vm=(vmat)mem_get(sizeof(vmat_struct),"vmat_create");
    for (i=0;i<4;i++) { for (j=0;j<4;j++) vm->m[i][j]=0.0; }
    return vm;
}

void vmat_destroy(vmat vm)
{
    mem_put((char *)vm,"vmat_destroy");
}

void vmat_ident(vmat vmr)
{
    vmat_scale(1.0,vmr);
}

void vmat_scale(double q,vmat vmr)
{
    int i,j;
    vmatmat m;

    m=vmr->m;
    for (i=0;i<4;i++) { for (j=0;j<4;j++) m[i][j]=(i==j?(i<3?q:1.0):0.0); }
}

void vmat_move(double dx,double dy,double dz,vmat vmr)
{
    int i,j;
    vmatmat m;

    m=vmr->m;
    for (i=0;i<4;i++) { for (j=0;j<4;j++) m[i][j]=(i==j?1.0:0.0); }
    m[0][3]=dx; m[1][3]=dy; m[2][3]=dz;
}

void vmat_xy(double a,vmat vmr)
{
    int i,j;
    double ca,sa;
    vmatmat m;

    m=vmr->m;
    for (i=0;i<4;i++) { for (j=0;j<4;j++) m[i][j]=(i==j?1.0:0.0); }
    ca=cos(a); sa=sin(a);
    m[0][0]=  ca; m[0][1]= -sa;
    m[1][0]=  sa; m[1][1]=  ca;
}

void vmat_yz(double a,vmat vmr)
{
    int i,j;
    double ca,sa;
    vmatmat m;

    m=vmr->m;
    for (i=0;i<4;i++) { for (j=0;j<4;j++) m[i][j]=(i==j?1.0:0.0); }
    ca=cos(a); sa=sin(a);
    m[1][1]=  ca; m[1][2]= -sa;
    m[2][1]=  sa; m[2][2]=  ca;
}

void vmat_zx(double a,vmat vmr)
{
    int i,j;
    double ca,sa;
    vmatmat m;

    m=vmr->m;
    for (i=0;i<4;i++) { for (j=0;j<4;j++) m[i][j]=(i==j?1.0:0.0); }
    ca=cos(a); sa=sin(a);
    m[2][2]=  ca; m[2][0]= -sa;
    m[0][2]=  sa; m[0][0]=  ca;
}

void vmat_per(double q,vmat vmr)
{
    int i,j;
    double z0,z1,z2;
    vmatmat m;

    m=vmr->m;
    for (i=0;i<4;i++) { for (j=0;j<4;j++) m[i][j]=(i==j?1.0:0.0); }
/**************
    m[3][2]= -1.0;
    m[3][3]=0.0;
    m[2][2]= -(200.0/99.99-1.0);
    m[2][3]= -(2.0/99.99);
**************/
    m[0][0]=q; m[1][1]=q;

    // original: z0=0.0; z1=z0+0.001; z2=z0+5.0*q;
    z0=0.0; z1=z0+0.001; z2=z0+5.0*q;
    // too tight: z0=0.0; z1=z0+0.1; z2=z0+1.5*q;
    // sort-of okay, but needs work: z0=0.0; z1=z0+0.01; z2=z0+4.0*q;

    m[3][2]=1.0;
    m[3][3]= -z0;
    m[2][2]=(2.0*z0-z1-z2)/(z1-z2);
    m[2][3]=z0-z1-m[2][2]*z1;
}

void vmat_mul(vmat vma,vmat vmb,vmat vmr)
{
    int i,j,k;
    double q;
    vmatmat ma,mb,mr;

    ma=vma->m; mb=vmb->m; mr=vmr->m;
    for (i=0;i<4;i++) {
        for (j=0;j<4;j++) {
            q=0.0;
            for (k=0;k<4;k++) q+=ma[i][k]*mb[k][j];
            mr[i][j]=q;
        }
    }
}

void vmat_invert(vmat vm,vmat vmr)
{
    int i,j;
    frame fr,fri;

    fr=frame_create(4,4); fri=frame_create(4,4);
    for (i=0;i<4;i++) {
        for (j=0;j<4;j++) {
            frame_setij(fr,i,j,vm->m[i][j]);
        }
    }
    frame_invert(fr,fri);
    for (i=0;i<4;i++) {
        for (j=0;j<4;j++) {
            vmr->m[i][j]=frame_getij(fri,i,j);
        }
    }
    frame_destroy(fr); frame_destroy(fri);
}

void vmat_copy(vmat vm,vmat vmr)
{
    int i,j;

    for (i=0;i<4;i++) {
        for (j=0;j<4;j++) {
            vmr->m[i][j]=vm->m[i][j];
        }
    }
}

void vmat_apply(vmat vm,double *p,double *pr)
{
    int i,j;
    double s,r[4],a[4];

    for (i=0;i<3;i++) a[i]=p[i];
    a[3]=1.0;
    for (i=0;i<4;i++) { s=0.0; for (j=0;j<4;j++) s+=vm->m[i][j]*a[j]; r[i]=s; }
    for (i=0;i<3;i++) pr[i]=r[i]/r[3];
}

void vmat_print(vmat vm,str name)
{
    int i,j;

    printf("=== %s ===\n",name);
    for (i=0;i<4;i++) {
        printf("  [");
        for (j=0;j<4;j++) {
            printf("%14g%s",vm->m[i][j],(j<3?" ":""));
        }
        printf("]\n");
    }
}

int  vmat_write(vmat vm,char *filename)
{
    int i,j;
    pack pk;
    char *buf,val[256];

    pk=pack_create("vmat");
    for (i=0;i<4;i++) {
        for (j=0;j<4;j++) {
            sprintf(val,"%.15g",vm->m[i][j]);
            pack_add(pk,pack_create(val));
        }
    }
    buf=pack_save(pk);
    i=fio_write(filename,buf,mem_size(buf)-1);
    pack_destroy(pk); mem_put((char *)buf,"vmat_write");
    return i;
}

vmat vmat_read(char *filename)
{
    int n;
    char *buf;
    pack pk;
    vmat vm;

    n=fio_size(filename); if (n<5) return (vmat)0;
    fio_setblock(0,n+1);
    buf=fio_read(filename);
    if (!buf) return (vmat)0;
    pk=pack_load(buf);
    mem_put(buf,"vmat_read");
    if (!pk) return (vmat)0;
    vm=vmat_evalpack(pk);
    pack_destroy(pk);
    return vm;
}

vmat vmat_evalpack(pack pk)
{
    int i,j,k;
    pack pkc;
    vmat vm,vmc,vm0;

    vm=vmat_create(); vm0=vmat_create();
    if (str_cmp(pk->description,"vmat")==0) {
        for (pkc=pk->pkc0,k=0;pkc && k<16;pkc=pkc->pkl1,k++) {
            i=k/4; j=k-4*i;
            sscanf(pkc->description,"%lf",&(vm->m[i][j]));
        }
        return vm;
    }
    if (str_cmp(pk->description,"mul")==0) {
        vmat_ident(vm);
        for (pkc=pk->pkc0;pkc;pkc=pkc->pkl1) {
            vmc=vmat_evalpack(pkc);
            vmat_copy(vm,vm0);
            vmat_mul(vmc,vm0,vm);
            vmat_destroy(vmc);
        }
        vmat_destroy(vm0);
        return vm;
    }
    vmat_ident(vm);
    return vm;
}


