
// view.l                                           view transform support
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// 06-07-93: wrote
// 03-22-94: _dragrot2
// 06-11-99: _dragrot1=old _dragrot, new _dragrot() mixes 1 and 2


lib_public {

    lib_include "../geowin.l";

    lib_struct {
        double x,y,z;                         // quick transformation data
        double ux,uy,uz;                      //   "
        double vx,vy,vz;                      //   "
        double wx,wy,wz;                      //   "
        double x0,y0,scale;                   //   "
        double dragx,dragy;                   // old mouse pos. for drag
        void (*event)(view);                  // callback event handler
        char *data;                           // application-specific data
    } view;

    typedef void (*viewevent)(view);

    view view_create(double x0,double y0,double scale,viewevent event);
    void view_destroy(view v);
    void view_resize(double x0,double y0,double scale);
    void view_set(view v);
    void view_wtoc(double x,double y,double z,
        double *ru,double *rv,double *rw);
    void view_wtos(double x,double y,double z,
        double *rx,double *ry,double *rz);
    void view_ctos(double u,double v,double w,
        double *rx,double *ry,double *rz);
    void view_stoc(double x,double y,double *ru,double *rw);
    void view_ctow(double u,double v,double w,
        double *rx,double *ry,double *rz);
    void view_startdragrot(double x,double y);
    void view_dragrot(double x,double y);
    void view_dragrot1(double x,double y);
    void view_dragrot2(double x,double y);
    void view_depthcolor(int r,int g,int b,double z);
    void view_drawsphere();

};

lib_application "Viewtest/viewtest.l";


#ifndef PI
#define PI 3.1415926535897931
#endif

#define UX uvwxyz[0]
#define UY uvwxyz[1]
#define UZ uvwxyz[2]
#define VX uvwxyz[3]
#define VY uvwxyz[4]
#define VZ uvwxyz[5]
#define WX uvwxyz[6]
#define WY uvwxyz[7]
#define WZ uvwxyz[8]

#define NORM(X,Y,Z,R) {                                                       \
    R=sqrt(X*X+Y*Y+Z*Z);                                                      \
    if (R<1e-12) { X=1.0; Y=0.0; Z=0.0; }                                     \
    else { X/=R; Y/=R; Z/=R; }                                                \
}

#define CROSS(BX,BY,BZ,CX,CY,CZ,AX,AY,AZ) {                                   \
    AX=BY*CZ-BZ*CY;                                                           \
    AY=BZ*CX-BX*CZ;                                                           \
    AZ=BX*CY-BY*CX;                                                           \
}

view view_v0;
int view_dragrotmode;

/*
 * sphere curves
 *   8 per axial direction
 *   projected from cube faces to sphere
 *   3 per face, common edges
 */

#define VIEW_SPHERE_CURVES   24
#define VIEW_SPHERE_SAMPLES   5

lib_struct {
    double x,y,z;
    int r,g,b;
} viewspt;

viewspt_struct viewspts_store[VIEW_SPHERE_CURVES][VIEW_SPHERE_SAMPLES];
viewspt viewspts[VIEW_SPHERE_CURVES][VIEW_SPHERE_SAMPLES];

void viewspt_fix(viewspt p);
void view_dragrot1sub(view v0,double x,double y,double *uvwxyz);
void view_dragrot2sub(view v0,double x,double y,double *uvwxyz);


void viewspt_fix(viewspt p)
{
    double q;

    p->r=0; p->g=0; p->b=0;
    if (p->x< -0.99) { p->g++; p->b++; }
    if (p->x>0.99) p->r++;
    if (p->y< -0.99) { p->r++; p->b++; }
    if (p->y>0.99) p->g++;
    if (p->z< -0.99) { p->r++; p->g++; }
    if (p->z>0.99) p->b++;
    p->r=p->r?255:0; p->g=p->g?255:0; p->b=p->b?255:0;
    q=sqrt(p->x*p->x+p->y*p->y+p->z*p->z);
    p->x/=q; p->y/=q; p->z/=q;
}


lib_init {
    view_v0=(view)0;
    view_dragrotmode=0;

    /*
     * curves on sphere
     */
    {
        int i,j,k,n;
        double u,v,w;
        viewspt *p0,*p1,*p2,p;

        for (i=0;i<VIEW_SPHERE_CURVES;i++) {
            for (j=0;j<VIEW_SPHERE_SAMPLES;j++) {
                viewspts[i][j]= &viewspts_store[i][j];
            }
        }
        n=0;
        for (i=0;i<3;i++) {
            for (j=0;j<3;j++) {
                if (i!=1 || j!=1) {
                    u=(double)(i-1); v=(double)(j-1);
                    p0=viewspts[n]; p1=viewspts[n+1]; p2=viewspts[n+2];
                    for (k=0;k<VIEW_SPHERE_SAMPLES;k++) {
                        w=2.0*(double)k/(double)(VIEW_SPHERE_SAMPLES-1)-1.0;
                        p=p0[k]; p->x=u; p->y=v; p->z=w; viewspt_fix(p);
                        p=p1[k]; p->x=v; p->y=w; p->z=u; viewspt_fix(p);
                        p=p2[k]; p->x=w; p->y=u; p->z=v; viewspt_fix(p);
                    }
                    n+=3;
                }
            }
        }
    }

};


view view_create(double x0,double y0,double scale,viewevent event)
{
    view v;

    v=(view)mem_get(sizeof(view_struct),"view_create");
    v->x0=x0; v->y0=y0; v->scale=scale;
    v->x=0.0; v->y=0.0; v->z=0.0;
    v->ux=1.0; v->uy=0.0; v->uz=0.0;
    v->vx=0.0; v->vy=1.0; v->vz=0.0;
    v->wx=0.0; v->wy=0.0; v->wz=1.0;
    v->dragx=v->dragy= -1.0;
    v->event=event; v->data=(char *)0;
    view_set(v);
    return v;
}

void view_destroy(view v)
{
    if (v==view_v0) view_v0=(view)0;
    mem_put((char *)v,"view_destroy");
}

void view_resize(double x0,double y0,double scale)
{
    view_v0->x0=x0; view_v0->y0=y0; view_v0->scale=scale;
}

void view_set(view v)
{
    view_v0=v;
}


/*
 * ===== coordinate transforms =====
 */
 
/*
 * world to camera
 */

void view_wtoc(double x,double y,double z,double *ru,double *rv,double *rw)
{
    double x0,y0,z0;
    view v0;

    v0=view_v0;
    x0=x-v0->x; y0=y-v0->y; z0=z-v0->z;
    *ru=v0->ux*x0+v0->uy*y0+v0->uz*z0;
    *rv=v0->vx*x0+v0->vy*y0+v0->vz*z0;
    *rw=v0->wx*x0+v0->wy*y0+v0->wz*z0;
}
 
/*
 * world to screen
 */

void view_wtos(double x,double y,double z,double *rx,double *ry,double *rz)
{
    double u,v,w;

    view_wtoc(x,y,z,&u,&v,&w);
    view_ctos(u,v,w,rx,ry,rz);
}
 
/*
 * camera to screen
 */

void view_ctos(double u,double v,double w,double *rx,double *ry,double *rz)
{
    *rx=view_v0->x0+view_v0->scale*u;
    *ry=view_v0->y0+view_v0->scale*w;
    *rz=v;
}
 
/*
 * screen to camera
 */

void view_stoc(double x,double y,double *ru,double *rw)
{
    *ru=(x-view_v0->x0)/view_v0->scale;
    *rw=(y-view_v0->y0)/view_v0->scale;
}
 
/*
 * camera to world
 */

void view_ctow(double u,double v,double w,double *rx,double *ry,double *rz)
{
    view v0;

    v0=view_v0;
    *rx=v0->ux*u+v0->vx*v+v0->wx*w+v0->x;
    *ry=v0->uy*u+v0->vy*v+v0->wy*w+v0->y;
    *rz=v0->uz*u+v0->vz*v+v0->wz*w+v0->z;
}

/*
 * ===== drag rotation =====
 */

void view_startdragrot(double x,double y)
{
    view_v0->dragx=x; view_v0->dragy=y;
}

#define R0 0.3
#define R1 0.6

void view_dragrot(double x,double y)
{
    double uvwxyz[9];
    double uvwxyz1[9];
    double uvwxyz2[9];
    view v0;

    v0=view_v0;
    if (x==v0->dragx && y==v0->dragy) return;

    switch (view_dragrotmode) {
    case 0:
        view_dragrot1sub(v0,x,y,uvwxyz1);
        view_dragrot2sub(v0,x,y,uvwxyz2);

        {
            int i;
            double x1,y1,r,a;

            x1=(x-v0->x0)/v0->scale; y1=(y-v0->y0)/v0->scale;
            r=sqrt(x1*x1+y1*y1);
            a=(r-R0)/(R1-R0); if (a<0.0) a=0.0; if (a>1.0) a=1.0;
            for (i=0;i<9;i++) uvwxyz[i]=(1.0-a)*uvwxyz2[i]+a*uvwxyz1[i];
        }

        break;
    case 1:
        view_dragrot1sub(v0,x,y,uvwxyz);
        break;
    case 2:
    default:
        view_dragrot2sub(v0,x,y,uvwxyz);
        break;
    }

    {
        double r;

        CROSS(UX,UY,UZ,VX,VY,VZ,WX,WY,WZ)
        NORM(WX,WY,WZ,r)
        CROSS(VX,VY,VZ,WX,WY,WZ,UX,UY,UZ)
        NORM(UX,UY,UZ,r)
        CROSS(WX,WY,WZ,UX,UY,UZ,VX,VY,VZ)
        NORM(VX,VY,VZ,r)
    }
    v0->ux=UX; v0->uy=UY; v0->uz=UZ;
    v0->vx=VX; v0->vy=VY; v0->vz=VZ;
    v0->wx=WX; v0->wy=WY; v0->wz=WZ;
    v0->dragx=x; v0->dragy=y;
    (*v0->event)(v0);
}

void view_dragrot1(double x,double y)
{
    int m0;

    m0=view_dragrotmode; view_dragrotmode=1;
    view_dragrot(x,y);
    view_dragrotmode=m0;
}

void view_dragrot2(double x,double y)
{
    int m0;

    m0=view_dragrotmode; view_dragrotmode=2;
    view_dragrot(x,y);
    view_dragrotmode=m0;
}

void view_dragrot1sub(view v0,double x,double y,double *uvwxyz)
{
    double xc,yc,r,a,b,ct,st;
    double x0,y0,a0,r0;
    double x1,y1,a1,r1;
    double ux,uy,uz;
    double vx,vy,vz;
    double wx,wy,wz;

    /*
     * get old and new mouse positions, polar coordinates
     */
    xc=v0->x0; yc=v0->y0;
    r=v0->scale;
    x0=(v0->dragx-xc)/r; y0=(v0->dragy-yc)/r;
    r0=sqrt(x0*x0+y0*y0); a0=(r0>1e-6)?atan2(y0,x0):0.0;
    x1=(x-xc)/r; y1=(y-yc)/r;
    r1=sqrt(x1*x1+y1*y1); a1=(r1>1e-6)?atan2(y1,x1):0.0;
    b=0.5*PI*(r1-r0);

    UX=v0->ux; UY=v0->uy; UZ=v0->uz;
    VX=v0->vx; VY=v0->vy; VZ=v0->vz;
    WX=v0->wx; WY=v0->wy; WZ=v0->wz;

    /*
     * radial rotation (generalized pan)
     */
    ct=cos(-a0); st=sin(-a0);
    ux=ct*UX-st*WX; wx=st*UX+ct*WX;
    uy=ct*UY-st*WY; wy=st*UY+ct*WY;
    uz=ct*UZ-st*WZ; wz=st*UZ+ct*WZ;
    UX=ux; WX=wx;
    UY=uy; WY=wy;
    UZ=uz; WZ=wz;

    ct=cos(b); st=sin(b);
    ux=ct*UX-st*VX; vx=st*UX+ct*VX;
    uy=ct*UY-st*VY; vy=st*UY+ct*VY;
    uz=ct*UZ-st*VZ; vz=st*UZ+ct*VZ;
    UX=ux; VX=vx;
    UY=uy; VY=vy;
    UZ=uz; VZ=vz;

    ct=cos(a0); st=sin(a0);
    ux=ct*UX-st*WX; wx=st*UX+ct*WX;
    uy=ct*UY-st*WY; wy=st*UY+ct*WY;
    uz=ct*UZ-st*WZ; wz=st*UZ+ct*WZ;
    UX=ux; WX=wx;
    UY=uy; WY=wy;
    UZ=uz; WZ=wz;

    /*
     * axial rotation (twist)
     */
    ct=cos(a1-a0); st=sin(a1-a0);
    ux=ct*UX-st*WX; wx=st*UX+ct*WX;
    uy=ct*UY-st*WY; wy=st*UY+ct*WY;
    uz=ct*UZ-st*WZ; wz=st*UZ+ct*WZ;
    UX=ux; WX=wx;
    UY=uy; WY=wy;
    UZ=uz; WZ=wz;

    /*
     * fix up u,v,w
     */
/*
    ...more: u=NORM(CROSS(v,w))
    ...more: v=NORM(CROSS(w,u))
    ...more: w=NORM(CROSS(u,v))
*/
}

void view_dragrot2sub(view v0,double x,double y,double *uvwxyz)
{
    double xc,yc,r,a,ct,st;
    double au,av,aw,ar;
    double bu,bv,bw,br;
    double pu,pv,pw,pr;
    double qu,qv,qw,qr;
    double xu,xv,xw,xr;
    double yu,yv,yw,yr;
    double zu,zv,zw,zr;
    double ax,ay,az;
    double px,py,pz;
    double qx,qy,qz;
    double ax1,ay1,az1;
    double px1,py1,pz1;
    double qx1,qy1,qz1;

    UX=v0->ux; UY=v0->uy; UZ=v0->uz;
    VX=v0->vx; VY=v0->vy; VZ=v0->vz;
    WX=v0->wx; WY=v0->wy; WZ=v0->wz;

    /*
     * get unit old and new directions, normal and tangent frame
     */
    xc=v0->x0; yc=v0->y0;
    r=v0->scale;
    au=(v0->dragx-xc)/r; av= -1.0; aw=(v0->dragy-yc)/r;
    NORM(au,av,aw,ar);
    bu=(x-xc)/r; bv= -1.0; bw=(y-yc)/r;
    NORM(bu,bv,bw,br);
    CROSS(bu,bv,bw,au,av,aw,pu,pv,pw);
    NORM(pu,pv,pw,pr);
    CROSS(au,av,aw,pu,pv,pw,qu,qv,qw);
    NORM(qu,qv,qw,qr);
    a=acos(au*bu+av*bv+aw*bw);

    /*
     * get world frame in normal/tangent frame
     */
    xu=UX; xv=VX; xw=WX;
    yu=UY; yv=VY; yw=WY;
    zu=UZ; zv=VZ; zw=WZ;
    ax=au*xu+av*xv+aw*xw; px=pu*xu+pv*xv+pw*xw; qx=qu*xu+qv*xv+qw*xw;
    ay=au*yu+av*yv+aw*yw; py=pu*yu+pv*yv+pw*yw; qy=qu*yu+qv*yv+qw*yw;
    az=au*zu+av*zv+aw*zw; pz=pu*zu+pv*zv+pw*zw; qz=qu*zu+qv*zv+qw*zw;

    /*
     * rotate world frame from a to b in aq plane
     */
    ct=cos(a); st=sin(a);
    ax1=ct*ax-st*qx; qx1=st*ax+ct*qx;
    ay1=ct*ay-st*qy; qy1=st*ay+ct*qy;
    az1=ct*az-st*qz; qz1=st*az+ct*qz;
    ax=ax1; qx=qx1;
    ay=ay1; qy=qy1;
    az=az1; qz=qz1;

    /*
     * get world frame back in camera frame
     */
    xu=ax*au+px*pu+qx*qu; xv=ax*av+px*pv+qx*qv; xw=ax*aw+px*pw+qx*qw;
    yu=ay*au+py*pu+qy*qu; yv=ay*av+py*pv+qy*qv; yw=ay*aw+py*pw+qy*qw;
    zu=az*au+pz*pu+qz*qu; zv=az*av+pz*pv+qz*qv; zw=az*aw+pz*pw+qz*qw;

    UX=xu; VX=xv; WX=xw;
    UY=yu; VY=yv; WY=yw;
    UZ=zu; VZ=zv; WZ=zw;
}

void view_depthcolor(int r,int g,int b,double z)
{
    int n;

    n=(int)floor(127.99*(1.0-z)+0.00001);
    if (n<0) n=0; if (n>255) n=255;
    win_color((r*n)/255,(g*n)/255,(b*n)/255);
}

void view_drawsphere()
{
    int i,k,n,sx0,sy0,sx,sy;
    viewspt *p0,p;
    double x,y,z;

    for (k=0;k<VIEW_SPHERE_CURVES;k++) {
        p0=viewspts[k];
        for (i=0;i<VIEW_SPHERE_SAMPLES;i++) {
            p=p0[i];
            view_wtos(p->x,p->y,p->z,&x,&y,&z);
            view_depthcolor(p->r,p->g,p->b,z);
            sx=(int)floor(x+0.5); sy=(int)floor(y+0.5);
            if (i) win_line(sx0,sy0,sx,sy);
            sx0=sx; sy0=sy;
        }
    }
}


