
lib_public {

    // qscene.l                             top-level ROAMing surfaces module
    // by Mark Duchaineau
    //
    // 11-30-99: wrote full draft code
    // 12-15-99: debugged and tested
    // 12-17-99: added error-radius trees

    lib_include "Win/win.l";
    lib_include "Tile/tile.l";


    // ->qstate bit fields to aggregate into the values below
    // NOTE: _ISOUT and _ISMAX are only used for leaves
    // ------------------------------------------------
    #define QSTATE_ISCREATE     0x001 /* qcell just created             */
    #define QSTATE_ISLEAF       0x020 /* 1=leaf, 0=non-leaf             */
    #define QSTATE_ISDRAW       0x010 /* 1=draw, 0=don't draw           */
    #define QSTATE_ISQUEUED     0x002 /* 1=on queue, 0=not queued       */
    #define QSTATE_ISOUT        0x004 /* 1=OUT, 0=not OUT               */
    #define QSTATE_ISMAX        0x008 /* 1=l at lmax, 0=l<lmax          */

    // ->qstate values (aggregates of bit fields above)
    // ------------------------------------------------
    #define QSTATE_CREATE       0x001 /* new qcell                      */
    #define QSTATE_LEAF_QUEUED  0x032 /* leaf in splitq (<lmax,!OUT)    */
    #define QSTATE_LEAF_MAX     0x038 /* leaf at lmax and !OUT          */
    #define QSTATE_LEAF_OUT     0x024 /* leaf OUT                       */
    #define QSTATE_MERGE_QUEUED 0x002 /* mergable non-leaf in mergeq    */
    #define QSTATE_MERGE_NOT    0x000 /* non-mergable non-leaf          */


    // number of buckets in a priority queue
    #define Q_IQMAX 0x1000


    lib_struct {                   // optimizer globals
        qtile qt0,qt1;             // list of tiles
        int ntri,ntrimax;          // current and target output triangle counts
        qpool qqp;                 // pool for fast qcell allocation
        qcell splitq[Q_IQMAX];     // split-queue buckets
        int iqsmax;                //   max occupied-bucket index
        qcell mergeq[Q_IQMAX];     // merge-queue buckets
        int iqmmin;                //   min occupied-bucket index
        int drawmode;              // 0=normal, 1=show mesh
        GLuint linetex_id;         // line-mask texture-object id
        float lineuv[3][2];        // line-mask (u,v) texture coords
        int debugsplit;            // on-shot trigger debug prints for optimize
    } qscene;

    lib_struct {                   // base-mesh quad array
        qscene qs;                 // scene that contains this tile
        qtile qt0,qt1;             // links in qs's tile list
        qtile qt[4];               // neighboring tiles
        int qt_ei[4];              // touching-edge index in neighboring tile
        int lmax;                  // q->l at finest resolution (==ns*2)
        int ns;                    // p array size (1<<ns)+1 x (1<<ns)+1
        float *p;                  // surface points stored (x,y,z,x,y,z...)
        float *e;                  // error radius per bintree triangle
        int tex_size;              // texture size tex_size*tex_size
        int *tex_ras;              // texture raster image (OpenGL RGBA format)
        GLuint tex_id;             // texture-object id from glGenTextures()
        qcell q0,q1;               // root bintree triangles
    } qtile;

    lib_struct {                   // active bintree triangle
        qtile qt;                  // qtile containing this qcell
        int l;                     // level in the bintree (0=top)
        int qstate;                // queueing status bits (see defs above)
        int bi;                    // bintree index (for error-radius lookup)
        int ij[3][2];              // vertex indices in qt->p array
        float p[3][3];             // triangle vertex (x,y,z) positions
        float uv[3][2];            // triangle vertex (u,v) texture coords
        float e;                   // error from qt->e[bi]
        int iq;                    // priority (queue bucket index)
        qcell qp,qc0,qc1;          // parent and kids
        qcell q0,q1,qb;            // neighbors (left, right, base)
        qcell qq0,qq1;             // queue links
    } qcell;


    qscene qscene_create();
    void qscene_readqtile(qscene qs,char *filename);
    // ...more: setfrustum() call
    void qscene_optimize(qscene qs);
    void qscene_drawmode(qscene qs,int mode);
    void qscene_draw(qscene qs);

};


lib_application "Tile2qtile/tile2qtile.l";
lib_application "Qtilespin/qtilespin.l";


void qscene_flush(qscene qs);
void qscene_drawsub(qcell q);
void qscene_drawsubline(qcell q);
qtile qtile_create(qscene qs);
qcell qcell_create(qtile qt,int l,int ij[][2],int bi);
void qcell_queue(qcell q,int qstate,int iq);
void qcell_split(qcell q);
int qcell_priority(qcell q);


qscene qscene_create()
{
    int iq;
    qscene qs;

    qs=(qscene)mem_get(sizeof(qscene_struct),"qscene_create");
    qs->qt0=(qtile)0; qs->qt1=(qtile)0;
    qs->qqp=(qpool)0;
    qs->ntrimax=3000;
    qs->drawmode=0;
    qs->debugsplit=0;

    // generate line-mask texture for triangulation display
    {
        int size,x,y;
        char *ras,*rp;
        float a,a0,a1,a2,b,b2;

        #define AA 10.0

        size=128;
        a0=0.5/(double)size; a1=1.0-a0;
        qs->lineuv[0][0]=a1; qs->lineuv[0][1]=a0;
        qs->lineuv[1][0]=a0; qs->lineuv[1][1]=a1;
        qs->lineuv[2][0]=a0; qs->lineuv[2][1]=a0;
        ras=rp=mem_get(4*size*size,"qscene_create");
        a=0.01; b=1.0/sqrt(2.0); b2=1.0/((AA-1.0)*a);
        for (y=0;y<size;y++) {
            for (x=0;x<size;x++) {
                *rp++ =60; *rp++ =40; *rp++ =10;
                //*rp++ =((x<5 || y<5 || x+y>size-7)?255:0);
                a0=(double)x/(double)(size-1);
                a1=(double)y/(double)(size-1);
                a2=b*(1.0-(a0+a1));
                if (a0<a) a0=0.0;
                else { if (a0<AA*a) a0=(a0-a)*b2; else a0=1.0; }
                if (a1<a) a1=0.0;
                else { if (a1<AA*a) a1=(a1-a)*b2; else a1=1.0; }
                if (a2<a) a2=0.0;
                else { if (a2<AA*a) a2=(a2-a)*b2; else a2=1.0; }
                *rp++ =(int)floor(255.99999*(1.0-a0*a1*a2));
            }
        }
        glGenTextures(1,&(qs->linetex_id));
        glBindTexture(GL_TEXTURE_2D,qs->linetex_id);
        glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,
            size,size,0,GL_RGBA,
            GL_UNSIGNED_BYTE,(GLvoid *)ras);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
        mem_put(ras,"qscene_create");
    }

    qscene_flush(qs);
    return qs;
}

void qscene_readqtile(qscene qs,char *filename)
{
    int *buf,*bp,n,ns,np,tex_size,ei,ti;
    GLuint *tex_ids;
    qtile qt,*qttab;

    // NOTE: don't de-allocate buf!
    // tiles point into it to avoid allocs and copies
    buf=(int *)fio_read(filename);
    if (!buf) { fprintf(stderr,"can't read %s\n",filename); exit(1); }

    n=buf[0]; ns=buf[1]; tex_size=buf[2]; 
    qttab=(qtile *)mem_get(n*sizeof(qtile),"qscene_readqtile");
    tex_ids=(GLuint *)mem_get(n*sizeof(GLuint),"qscene_readqtile");
    glGenTextures(n,tex_ids);
    for (ti=0;ti<n;ti++) qttab[ti]=qt=qtile_create(qs);
    np=(1<<ns)+1;
    for (ti=0,bp=buf+3;ti<n;ti++) {
        qt=qttab[ti];
        qt->lmax=ns<<1; qt->ns=ns;

        for (ei=0;ei<4;ei++,bp+=2)
            { qt->qt[ei]=qttab[bp[0]]; qt->qt_ei[ei]=bp[1]; }

        qt->p=(float *)bp; bp+=3*np*np;

        qt->tex_size=tex_size;
        qt->tex_ras=bp; bp+=tex_size*tex_size;
        qt->tex_id=tex_ids[ti];
        glBindTexture(GL_TEXTURE_2D,qt->tex_id);
        glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,
            tex_size,tex_size,0,GL_RGBA,
            GL_UNSIGNED_BYTE,(GLvoid *)qt->tex_ras);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);

        qt->e=(float *)bp; bp+=2<<qt->lmax;

        qt->q0=(qcell)0; qt->q1=(qcell)0;
    }
    mem_put((char *)qttab,"qscene_readqtile");
    mem_put((char *)tex_ids,"qscene_readqtile");

    // put all tiles into clean base-mesh state
    qscene_flush(qs);
}

void qscene_optimize(qscene qs)
{
    int k,ij0[3][2],ij1[3][2];
    qcell q0,q1;
    qtile qt;

    // clear out old qcells
    qscene_flush(qs);

    // iteratively split the highest-priority splitq element
    // until ntri>=ntrimax
    while (qs->ntri<qs->ntrimax) qcell_split(qs->splitq[qs->iqsmax]);

    qs->debugsplit=0;
}

void qscene_drawmode(qscene qs,int mode)
{
    qs->drawmode=mode;
}

void qscene_draw(qscene qs)
{
    qtile qt;

    if (!qs->qqp) return;

    glEnable(GL_TEXTURE_2D);

    for (qt=qs->qt0;qt;qt=qt->qt1) {
        glBindTexture(GL_TEXTURE_2D,qt->tex_id);
        qscene_drawsub(qt->q0); qscene_drawsub(qt->q1);
    }

    if (qs->drawmode==1) {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); 
        glBindTexture(GL_TEXTURE_2D,qs->linetex_id);
        for (qt=qs->qt0;qt;qt=qt->qt1) {
            qscene_drawsubline(qt->q0); qscene_drawsubline(qt->q1);
        }
        glDisable(GL_BLEND);
    }

    glDisable(GL_TEXTURE_2D);
}


// ========== internal routines ==========

// destroy pool, remove tile root cells, clear buckets, ntri=0

void qscene_flush(qscene qs)
{
    int iq;
    qcell *splitq,*mergeq;
    qtile qt;

    // faster to not do de-allocate/allocate on pool storage, but to
    // just put all pool elements in the free state
    if (qs->qqp) qpool_destroy(qs->qqp);
    qs->qqp=(qpool)0;

    for (qt=qs->qt0;qt;qt=qt->qt1) { qt->q0=(qcell)0; qt->q1=(qcell)0; }
    splitq=qs->splitq; mergeq=qs->mergeq;
    for (iq=0;iq<Q_IQMAX;iq++) { splitq[iq]=mergeq[iq]=(qcell)0; }
    qs->iqsmax= -1;
    qs->iqmmin=Q_IQMAX;
    qs->ntri=0;

    // initialize to base-mesh state
    if (qs->qt0) {
        int k,ij0[3][2],ij1[3][2];
        qcell q0,q1;

        // create pool and tile root cells, queue the root cells
        qs->qqp=qpool_create(sizeof(qcell_struct));
        // NOTE: we assume here that all qt's have the same ns
        k=1<<qs->qt0->ns;
        ij0[0][0]=0; ij0[0][1]=0;
        ij0[1][0]=k; ij0[1][1]=k;
        ij0[2][0]=0; ij0[2][1]=k;
        ij1[0][0]=k; ij1[0][1]=k;
        ij1[1][0]=0; ij1[1][1]=0;
        ij1[2][0]=k; ij1[2][1]=0;
        for (qt=qs->qt0;qt;qt=qt->qt1) {
            qt->q0=qcell_create(qt,0,ij0,2);
            qcell_queue(qt->q0,QSTATE_LEAF_QUEUED,qcell_priority(qt->q0));
            qt->q1=qcell_create(qt,0,ij1,3);
            qcell_queue(qt->q1,QSTATE_LEAF_QUEUED,qcell_priority(qt->q1));
        }
        // hook them all up
        for (qt=qs->qt0;qt;qt=qt->qt1) {
            q0=qt->q0; q1=qt->q1;
            q0->qb=q1; q1->qb=q0;
            q1->q1=(qt->qt_ei[0]<2?qt->qt[0]->q1:qt->qt[0]->q0);
            q1->q0=(qt->qt_ei[1]<2?qt->qt[1]->q1:qt->qt[1]->q0);
            q0->q1=(qt->qt_ei[2]<2?qt->qt[2]->q1:qt->qt[2]->q0);
            q0->q0=(qt->qt_ei[3]<2?qt->qt[3]->q1:qt->qt[3]->q0);
        }
    }
}

void qscene_drawsub(qcell q)
{
    int vi;

    if (q->qc0) { qscene_drawsub(q->qc0); qscene_drawsub(q->qc1); return; }
    glBegin(GL_TRIANGLE_STRIP);
    for (vi=0;vi<3;vi++)
        { glTexCoord2fv(q->uv[vi]); glVertex3fv(q->p[vi]); }
    glEnd();
}

void qscene_drawsubline(qcell q)
{
    int vi;

    if (q->qc0)
        { qscene_drawsubline(q->qc0); qscene_drawsubline(q->qc1); return; }
    glBegin(GL_TRIANGLE_STRIP);
    for (vi=0;vi<3;vi++)
        { glTexCoord2fv(q->qt->qs->lineuv[vi]); glVertex3fv(q->p[vi]); }
    glEnd();
}

qtile qtile_create(qscene qs)
{
    qtile qt;

    qt=(qtile)mem_get(sizeof(qtile_struct),"qtile_create");
    qt->qs=qs;
    qt->qt0=qs->qt1; qt->qt1=(qtile)0;
    if (qt->qt0) qt->qt0->qt1=qt; else qs->qt0=qt;
    if (qt->qt1) qt->qt1->qt0=qt; else qs->qt1=qt;
    return qt;
}

qcell qcell_create(qtile qt,int l,int ij[][2],int bi)
{
    int vi,i,j,dj,k;
    qscene qs;
    qcell q;
    float *pp;
    double tc0,tc1;

    qs=qt->qs;
    q=(qcell)qpool_get(qs->qqp);
    q->qt=qt;
    q->l=l;
    q->bi=bi;
    q->qstate=QSTATE_CREATE;
    dj=(1<<qt->ns)+1;
    tc0=0.5/(double)qt->tex_size;
    tc1=(double)(qt->tex_size-1)/(double)(qt->tex_size*(dj-1));
    for (vi=0;vi<3;vi++) {
        q->ij[vi][0]=i=ij[vi][0];
        q->ij[vi][1]=j=ij[vi][1];
        pp=qt->p+3*(j*dj+i);
        for (k=0;k<3;k++) q->p[vi][k]=pp[k];
        q->uv[vi][0]=tc0+tc1*(double)i;
        q->uv[vi][1]=tc0+tc1*(double)j;
    }
    q->e=(l<qt->lmax?qt->e[bi]:0.0);
    q->iq= -1;
    q->qp=q->qc0=q->qc1=(qcell)0;
    q->q0=q->q1=q->qb=(qcell)0;
    q->qq0=q->qq1=(qcell)0;
    return q;
}

void qcell_queue(qcell q,int qstate,int iq)
{
    int qstate0,iq0,iqx;
    qcell qq0,qq1,*pq,*qtab;
    qtile qt;
    qscene qs;

    // get current info
    qt=q->qt; qs=qt->qs; qstate0=q->qstate; iq0=q->iq;

    // set to new values
    q->qstate=qstate; q->iq=iq;

    // decrease triangle count if q was contributing
    if (qstate0&QSTATE_ISDRAW) qs->ntri--;

    // remove from old queue list, if any
    if (qstate0&QSTATE_ISQUEUED) {
        if (qstate0&QSTATE_ISLEAF) pq=qs->splitq+iq0;
        else pq=qs->mergeq+iq0;
        qq0=q->qq0; qq1=q->qq1;
        if (qq0) qq0->qq1=qq1; else *pq=qq1;
        if (qq1) qq1->qq0=qq0;
        if (!qq0 && !qq1) {
            if (qstate0&QSTATE_ISLEAF) {
                if (iq0==qs->iqsmax) {
                    qtab=qs->splitq;
                    if (!qtab[0]) qtab[0]=(qcell)1;
                    for (iqx=qs->iqsmax;!qtab[iqx];iqx--) ;
                    if (qtab[0]==(qcell)1) {
                        if (iqx==0) iqx= -1;
                        qtab[0]=(qcell)0;
                    }
                    qs->iqsmax=iqx;
                }
            }else{
                if (iq0==qs->iqmmin) {
                    qtab=qs->mergeq;
                    if (!qtab[Q_IQMAX-1]) qtab[Q_IQMAX-1]=(qcell)1;
                    for (iqx=qs->iqmmin;!qtab[iqx];iqx++) ;
                    if (qtab[Q_IQMAX-1]==(qcell)1) {
                        if (iqx==Q_IQMAX-1) iqx=Q_IQMAX;
                        qtab[Q_IQMAX-1]=(qcell)0;
                    }
                    qs->iqmmin=iqx;
                }
            }
        }
    }

    // add to queue if needed
    if (qstate&QSTATE_ISQUEUED) {
        if (qstate&QSTATE_ISLEAF) {
            pq=qs->splitq+iq; if (iq>qs->iqsmax) qs->iqsmax=iq;
        }else{
            pq=qs->mergeq+iq; if (iq<qs->iqmmin) qs->iqmmin=iq;
        }
        qq1= *pq; *pq=q; q->qq0=(qcell)0; q->qq1=qq1; if (qq1) qq1->qq0=q;
    }

    // increase triangle count if q now contributes
    if (qstate&QSTATE_ISDRAW) qs->ntri++;

}

void qcell_split(qcell q)
{
    int l,ij[3][2],iq,iqb,k,bi;
    qtile qt,qtb;
    qcell qb,qc0,qc1,qbc0,qbc1,qx;

    if (q->qc0) return;

    qt=q->qt; qtb=q->qb->qt; l=q->l+1;
    if (q->qb->l<q->l) qcell_split(q->qb);
    qb=q->qb;
    iq=q->iq; iqb=qb->iq;

    // NOTE: this will become MERGE_QUEUED when the merge queue is needed
    qcell_queue(q,QSTATE_MERGE_NOT,0); qcell_queue(qb,QSTATE_MERGE_NOT,0);

    bi=q->bi<<1;
    ij[2][0]=(q->ij[0][0]+q->ij[1][0])>>1;
    ij[2][1]=(q->ij[0][1]+q->ij[1][1])>>1;
    ij[0][0]=q->ij[2][0]; ij[0][1]=q->ij[2][1];
    ij[1][0]=q->ij[0][0]; ij[1][1]=q->ij[0][1];
    qc0=qcell_create(qt,l,ij,bi);
    ij[0][0]=q->ij[1][0]; ij[0][1]=q->ij[1][1];
    ij[1][0]=q->ij[2][0]; ij[1][1]=q->ij[2][1];
    qc1=qcell_create(qt,l,ij,bi+1);
    bi=qb->bi<<1;
    ij[2][0]=(qb->ij[0][0]+qb->ij[1][0])>>1;
    ij[2][1]=(qb->ij[0][1]+qb->ij[1][1])>>1;
    ij[0][0]=qb->ij[2][0]; ij[0][1]=qb->ij[2][1];
    ij[1][0]=qb->ij[0][0]; ij[1][1]=qb->ij[0][1];
    qbc0=qcell_create(qtb,l,ij,bi);
    ij[0][0]=qb->ij[1][0]; ij[0][1]=qb->ij[1][1];
    ij[1][0]=qb->ij[2][0]; ij[1][1]=qb->ij[2][1];
    qbc1=qcell_create(qtb,l,ij,bi+1);

    if (qt->qs->debugsplit && qt->qs->ntrimax-qt->qs->ntri<20) {
        printf("split displacement: [%g,%g,%g]\n",
            qc0->p[2][0]-(q->p[0][0]+q->p[1][0])/2.0,
            qc0->p[2][1]-(q->p[0][1]+q->p[1][1])/2.0,
            qc0->p[2][2]-(q->p[0][2]+q->p[1][2])/2.0);
        printf("split priority: %g => %d/%d\n",q->e,iq,Q_IQMAX);
    }

    q->qc0=qc0; q->qc1=qc1; qb->qc0=qbc0; qb->qc1=qbc1;
    qc0->qp=q; qc1->qp=q; qbc0->qp=qb; qbc1->qp=qb;

    #define BASEATTACH(Q,QCX,QX) {                                            \
        qx=Q->QX; QCX->qb=qx;                                                 \
        if (qx->q0==Q) qx->q0=QCX;                                            \
        else { if (qx->q1==Q) qx->q1=QCX; else qx->qb=QCX; }                  \
    }

    BASEATTACH(q,qc0,q0)
    BASEATTACH(q,qc1,q1)
    BASEATTACH(qb,qbc0,q0)
    BASEATTACH(qb,qbc1,q1)
    qc0->q0=qc1; qc0->q1=qbc1;
    qc1->q0=qbc0; qc1->q1=qc0;
    qbc0->q0=qbc1; qbc0->q1=qc1;
    qbc1->q0=qc0; qbc1->q1=qbc0;

    // non-finest qcells go on split queue
    // unqueued state for finest-res qcells
    if (l==qt->lmax) {
        qcell_queue(qc0,QSTATE_LEAF_MAX,0);
        qcell_queue(qc1,QSTATE_LEAF_MAX,0);
        qcell_queue(qbc0,QSTATE_LEAF_MAX,0);
        qcell_queue(qbc1,QSTATE_LEAF_MAX,0);
    }else{
        qcell_queue(qc0,QSTATE_LEAF_QUEUED,qcell_priority(qc0));
        qcell_queue(qc1,QSTATE_LEAF_QUEUED,qcell_priority(qc1));
        qcell_queue(qbc0,QSTATE_LEAF_QUEUED,qcell_priority(qbc0));
        qcell_queue(qbc1,QSTATE_LEAF_QUEUED,qcell_priority(qbc1));
    }
}

int qcell_priority(qcell q)
{
    double e,emin,emax;

    e=q->e;
    emin=0.00001; emax=10.0;
    if (e<=emin) return 0;
    if (e>=emax) return Q_IQMAX-1;
    e=(log(e)-log(emin))/(log(emax)-log(emin));
    return (int)floor(0.999990*(double)Q_IQMAX*e+0.000005);
}


