
lib_public {

    // surf_ops.l                                     surface misc. support
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 03-31-95: split from main surf module
    // 03-30-96: edgestate op, moved _swapbody to surf.l
    // 03-30-96: edgestate op, moved _swapbody() to surf.l
    // 02-13-97: _copyto()
    // 08-19-97: "spkread" command
    // 09-28-99: splitb1 and splitb3 ops
    // 01-04-00: mesbread op
    // 10-04-00: _smooth() (moved from bof2mes.l)

    lib_include "../surf.l";

    surf surf_parse(pack pk);
    void surf_createfaces(surf s);
    void surf_createsharpfaces(surf s);
    surf surf_copy(surf s);
    void surf_copyto(surf s_src,surf s_dst);
    void surf_offset(surf s,double r);
    void surf_perturb(surf s,double r);
    void surf_smooth(surf s,int mode,float *atab);
    void surf_triangulate(surf s);
    void surf_simplify(surf s);
    void surf_clip(surf s,double x,double y,double z,
        double nx,double ny,double nz);
    void surf_hollow(surf s,double r);
    void surf_dual(surf s);
    void surf_bound(surf s,double *rxmin,double *rxmax,
        double *rymin,double *rymax,double *rzmin,double *rzmax);
    void surf_translate(surf s,double dx,double dy,double dz);
    void surf_scale(surf s,double q);
    void surf_sharpverts(surf s);
    void surf_sharpedges(surf s);
    void surf_unsharpverts(surf s);
    void surf_unsharpedges(surf s);
    void surf_sanitize(surf s);
    void surf_vertnorm(vert v,double *rnx,double *rny,double *rnz,double *rnq);
    void surf_getnorms(surf s,double **rnx,double **rny,double **rnz,
        double **rnq);
    void surf_putnorms(double *nx,double *ny,double *nz,double *nq);

};


surf surf_parsesub(pack pk);
void surf_setedgestates(surf s,str action);


surf surf_parse(pack pk)
{
    int r;
    surf s;

    r=rnd_cnt;
    rnd_cnt=0x31427586;
    s=surf_parsesub(pk);
    rnd_cnt=r;
    return s;
}

surf surf_parsesub(pack pk)
{
    int n;
    double r,h;
    pack pkc;
    str name;
    surf s;

    if (!pk) return surf_tetrahedron(1.0);
    if (str_cmp(pk->description,"tetrahedron")==0) {
        r=1.000; if (pk->pkc0) sscanf(pk->pkc0->description,"%lf",&r);
        return surf_tetrahedron(r);
    }
    if (str_cmp(pk->description,"cube")==0) {
        r=1.000; if (pk->pkc0) sscanf(pk->pkc0->description,"%lf",&r);
        return surf_cube(r);
    }
    if (str_cmp(pk->description,"octahedron")==0) {
        r=1.000; if (pk->pkc0) sscanf(pk->pkc0->description,"%lf",&r);
        return surf_octahedron(r);
    }
    if (str_cmp(pk->description,"dodecahedron")==0) {
        r=1.000; if (pk->pkc0) sscanf(pk->pkc0->description,"%lf",&r);
        return surf_dodecahedron(r);
    }
    if (str_cmp(pk->description,"icosohedron")==0) {
        r=1.000; if (pk->pkc0) sscanf(pk->pkc0->description,"%lf",&r);
        return surf_icosohedron(r);
    }
    if (str_cmp(pk->description,"cylinder")==0) {
        n=4; if (pkc=pk->pkc0) sscanf(pkc->description,"%d",&n);
        r=0.667; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&r);
        h=1.000; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&h);
        return surf_cylinder(n,r,h);
    }
    if (str_cmp(pk->description,"cone")==0) {
        n=4; if (pkc=pk->pkc0) sscanf(pkc->description,"%d",&n);
        r=0.667; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&r);
        h=1.000; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&h);
        return surf_cone(n,r,h);
    }
    if (str_cmp(pk->description,"unpack")==0) return surf_unpack(pk->pkc0);
    if (str_cmp(pk->description,"read")==0) {
        if (pk->pkc0) name=pk->pkc0->description; else name="input.mes";
        return surf_read(name);
    }
    if (str_cmp(pk->description,"mesbread")==0) {
        if (pk->pkc0) name=pk->pkc0->description; else name="input.mesb";
        return surf_mesbread(name);
    }
    if (str_cmp(pk->description,"macread")==0) {
        if (pk->pkc0) name=pk->pkc0->description; else name="input.obj";
        return surf_macread(name);
    }
    if (str_cmp(pk->description,"spkread")==0) {
        if (pk->pkc0) name=pk->pkc0->description; else name="input.spk";
        return surf_spkread(name);
    }
    if (str_cmp(pk->description,"hollow")==0) {
        if (pk->pkc0 && pk->pkc0!=pk->pkc1) {
            s=surf_parsesub(pk->pkc0);
            r=0.2; sscanf(pk->pkc1->description,"%lf",&r);
            surf_hollow(s,r); return s;
        }else return surf_tetrahedron(1.0);
    }
    if (str_cmp(pk->description,"simplify")==0)
        { s=surf_parsesub(pk->pkc0); surf_simplify(s); return s; }
    if (str_cmp(pk->description,"dual")==0)
        { s=surf_parsesub(pk->pkc0); surf_dual(s); return s; }
    if (str_cmp(pk->description,"split1")==0)
        { s=surf_parsesub(pk->pkc0); surf_split1(s); return s; }
    if (str_cmp(pk->description,"split2")==0)
        { s=surf_parsesub(pk->pkc0); surf_split2(s); return s; }
    if (str_cmp(pk->description,"split3")==0)
        { s=surf_parsesub(pk->pkc0); surf_split3(s); return s; }
    if (str_cmp(pk->description,"limit3")==0)
        { s=surf_parsesub(pk->pkc0); surf_split3limit(s); return s; }
    if (str_cmp(pk->description,"splitb1")==0)
        { s=surf_parsesub(pk->pkc0); surf_splitb1(s); return s; }
    if (str_cmp(pk->description,"splitb3")==0)
        { s=surf_parsesub(pk->pkc0); surf_splitb3(s); return s; }
    if (str_cmp(pk->description,"scale")==0) {
        s=surf_parsesub(pk->pkc0);
        r=1.0; if (pk->pkc1) sscanf(pk->pkc1->description,"%lf",&r);
        surf_scale(s,r); return s;
    }
    if (str_cmp(pk->description,"unsharp")==0) {
        s=surf_parsesub(pk->pkc0); surf_unsharpverts(s); surf_unsharpedges(s);
        return s;
    }
    if (str_cmp(pk->description,"edgestate")==0) {
        s=surf_parsesub(pk->pkc0);
        surf_setedgestates(s,(pk->pkc1?pk->pkc1->description:"0")); return s;
    }
    if (str_cmp(pk->description,"sharpverts")==0)
        { s=surf_parsesub(pk->pkc0); surf_sharpverts(s); return s; }
    if (str_cmp(pk->description,"sharpedges")==0)
        { s=surf_parsesub(pk->pkc0); surf_sharpedges(s); return s; }
    if (str_cmp(pk->description,"clip")==0) {
        double x,y,z,nx,ny,nz;

        pkc=pk->pkc0; s=surf_parsesub(pkc);
        x=0.0; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&x);
        y=0.0; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&y);
        z=0.0; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&z);
        nx=0.0; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&nx);
        ny=0.0; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&ny);
        nz=0.0; if (pkc && (pkc=pkc->pkl1)) sscanf(pkc->description,"%lf",&nz);
        surf_clip(s,x,y,z,nx,ny,nz); return s;
    }
    if (str_cmp(pk->description,"perturb")==0) {
        s=surf_parsesub(pk->pkc0);
        r=0.1; if (pk->pkc1) sscanf(pk->pkc1->description,"%lf",&r);
        surf_perturb(s,r); return s;
    }
    if (str_cmp(pk->description,"offset")==0) {
        s=surf_parsesub(pk->pkc0);
        r=0.05; if (pk->pkc1) sscanf(pk->pkc1->description,"%lf",&r);
        surf_offset(s,r); return s;
    }
    if (str_cmp(pk->description,"sanitize")==0)
        { s=surf_parsesub(pk->pkc0); surf_sanitize(s); return s; }
    return surf_tetrahedron(1.0);
}

void surf_createfaces(surf s)
{
    edge e;
    face f;

    BGNEDGES
        if (!e->f) f=face_create(e);
    ENDEDGES
}

void surf_createsharpfaces(surf s)
{
    edge e,e1;
    face f;

    BGNEDGES
        if (!e->f) {
            f=face_create(e);
            for (e=f->e0;e;e=e1) {
                e1=e->e->ve0; if (e1==f->e0) e1=(edge)0;
                edge_sharpen(e);
            }
        }
    ENDEDGES
}

surf surf_copy(surf s)
{
    surf s1;

    s1=surf_create();
    surf_copyto(s,s1);
    return s1;
}

void surf_copyto(surf s_src,surf s_dst)
{
    int vi,vi0,vi1,ei,nv,ne,nf;
    vert v,vx,*verts;
    edge e,e0,e1,*edges;
    face f;
    surf s,s1;

    // initial setup
    s=s_src; s1=s_dst;
    surf_size(s,&nv,&ne,&nf);
    verts=(vert *)mem_get(nv*sizeof(vert),"surf_copyto");
    edges=(edge *)mem_get(ne*sizeof(edge),"surf_copyto");

    // verts
    vi=0;
    BGNVERTS
        if (vert_issharp(v)) surf_sharp();
        verts[vi++]=vert_create(s1,v->x,v->y,v->z);
    ENDVERTS

    // edges
    ei=0;
    BGNEDGES
        vi0=surf_vtoi(e->v); vi1=surf_vtoi(e->e->v);
        if (vi0<vi1) {
            edge_enstate(edge_state(e));
            edges[ei]=edge_create(verts[vi0],verts[vi1]);
        }else edges[ei]=(edge)0;
        ei++;
    ENDEDGES
    ei=0;
    BGNEDGES
        if (!edges[ei]) edges[ei]=edges[surf_etoi(e->e)]->e;
        ei++;
    ENDEDGES

    // add edges to verts
    vi=0;
    BGNVERTS
        vx=verts[vi];
        for (e=e0=v->e0;e;e=e1) {
            e1=e->ve1; if (e1==e0) e1=(edge)0;
            vert_addedge(vx,edges[surf_etoi(e)]);
        }
        vi++;
    ENDVERTS

    // create faces
    ei=0;
    BGNEDGES
        if (e->f && !edges[ei]->f) f=face_create(edges[ei]);
        ei++;
    ENDEDGES

    // done with temporary index arrays
    mem_put((char *)verts,"surf_copyto");
    mem_put((char *)edges,"surf_copyto");
}

void surf_offset(surf s,double r)
{
    int vi;
    vert v;
    double *nx,*ny,*nz,*nq;

    surf_getnorms(s,&nx,&ny,&nz,&nq);
    vi=0;
    BGNVERTS v->x+=r*nx[vi]; v->y+=r*ny[vi]; v->z+=r*nz[vi]; vi++; ENDVERTS
    surf_putnorms(nx,ny,nz,nq);
}

void surf_perturb(surf s,double r)
{
    int vi;
    vert v;
    double *nx,*ny,*nz,*nq;
    double dx,dy,dz,q;

    surf_getnorms(s,&nx,&ny,&nz,&nq);
    vi=0;
    BGNVERTS
        //for (q=2.0;q>1.0;q=dx*dx+dy*dy+dz*dz)
        //    { dx=RND; dy=RND; dz=RND; }
        // v->x+=dx*r; v->y+=dy*r; v->z+=dz*r;
        q=r*RND;
        v->x+=q*nx[vi]; v->y+=q*ny[vi]; v->z+=q*nz[vi]; vi++;
    ENDVERTS
    surf_putnorms(nx,ny,nz,nq);
}

void surf_smooth(surf s,int mode,float *atab)
{
    vert v,v1;
    edge e,e0,e1;
    face f;
    int nv,ne,nf,vi,fi;
    double *xc,*yc,*zc,x,y,z,r,a,b,dx,dy,dz,dr;

    printf("  smooth...\n");
    surf_size(s,&nv,&ne,&nf);
    mem_noclear(); xc=(double *)mem_get(nf*sizeof(double),"smooth");
    mem_noclear(); yc=(double *)mem_get(nf*sizeof(double),"smooth");
    mem_noclear(); zc=(double *)mem_get(nf*sizeof(double),"smooth");
    fi=0;
    BGNFACES
        x=y=z=r=0.0;
        for (e=e0=f->e0;e;e=e1) {
            e1=e->e->ve0; if (e1==e0) e1=(edge)0;
            v=e->v; x+=v->x; y+=v->y; z+=v->z; r+=1.0;
        }
        r=1.0/r; xc[fi]=r*x; yc[fi]=r*y; zc[fi]=r*z;
        fi++;
    ENDFACES
    vi=0;
    BGNVERTS
        x=y=z=r=0.0;
        for (e=e0=v->e0;e;e=e1) {
            e1=e->ve1; if (e1==e0) e1=(edge)0;
            if (e->f) {
                fi=surf_ftoi(e->f);
                if (mode==1) {
                    dx=v->x-xc[fi]; dy=v->y-yc[fi]; dz=v->z-zc[fi];
                    dr=sqrt(dx*dx+dy*dy+dz*dz);
                }else dr=1.0;
                x+=dr*xc[fi]; y+=dr*yc[fi]; z+=dr*zc[fi];
                r+=dr;
            }else r= -1e6;
        }
        if (r>0.0 && !vert_issharp(v)) {
            r=1.0/r; x*=r; y*=r; z*=r;
            b=atab[vi]; a=1.0-b;
            v->x=b*v->x+a*x; v->y=b*v->y+a*y; v->z=b*v->z+a*z;
        }
        vi++;
    ENDVERTS
    mem_put((char *)xc,"smooth");
    mem_put((char *)yc,"smooth");
    mem_put((char *)zc,"smooth");
}

void surf_triangulate(surf s)
{
    // ...more: this should be code now in ecol_dotriangulate()
}

void surf_simplify(surf s)
{
    int n,n1,i,i1,first;
    double x,y,z,xc,yc,zc,q;
    vert v,v0,v1,*verts2;
    edge e,e0,e1,*edges,*edges1,*edges2;
    face f;

    BGNFACES
        first=1;
        do {
            n=0;
            for (e=e0=f->e0;e;e=e1) {
                e1=e->e->ve0; if (e1==e0) e1=(edge)0;
                n++;
            }
            if (n>5) {
                edges =(edge *)mem_get(n*sizeof(edge),"surf_simplify");
                edges1=(edge *)mem_get(n*sizeof(edge),"surf_simplify");
                x=y=z=0.0;
                for (e=e0=f->e0,i=0;e;e=e1,i++) {
                    e1=e->e->ve0; if (e1==e0) e1=(edge)0;
                    edges[i]=e; v=e->v; x+=v->x; y+=v->y; z+=v->z;
                }
                face_destroy(f);
                q=(double)n; xc=x/q; yc=y/q; zc=z/q;
                n1=(first?n:(n+1)>>1);
                for (q=1.0,i=n;i>5;i=(i+1)>>1,q+=1.0) ;
                if (first) q+=1.0;
                q=1.0/q;
                edges2=(edge *)mem_get(n1*sizeof(edge),"surf_simplify");
                verts2=(vert *)mem_get(n1*sizeof(vert),"surf_simplify");
                for (i1=0;i1<n1;i1++) {
                    i=(first?i1:i1<<1); v0=edges[i]->v;
                    v1=edges[(!first && i+1<n)?i+1:i]->v;
                    x=(1.0-q)*(v0->x+v1->x)/2.0+q*xc;
                    y=(1.0-q)*(v0->y+v1->y)/2.0+q*yc;
                    z=(1.0-q)*(v0->z+v1->z)/2.0+q*zc;
                    verts2[i1]=v=vert_create(s,x,y,z);
                    edges1[i]=edge_create(v0,v);
                    if (v1!=v0) edges1[i+1]=edge_create(v1,v);
                }
                for (i1=0;i1<n1;i1++) {
                    i=i1+1; if (i>=n1) i-=n1;
                    edges2[i1]=edge_create(verts2[i1],verts2[i]);
                }
                for (i=0;i<n;i++) vert_insertedge(edges1[i],edges[i],(edge)0);
                for (i1=0;i1<n1;i1++) {
                    i=(first?i1:i1<<1);
                    v=verts2[i1];
                    vert_addedge(v,edges2[i1]);
                    vert_addedge(v,edges2[i1-1<0?n1-1:i1-1]);
                    vert_addedge(v,edges1[i]);
                    if (!first && i+1<n) vert_addedge(v,edges1[i+1]);
                }
                for (i=0;i<n;i++) f=face_create(edges[i]);
                f=face_create(edges2[0]);
                mem_put((char *)edges,"surf_simplify");
                mem_put((char *)edges1,"surf_simplify");
                mem_put((char *)edges2,"surf_simplify");
                mem_put((char *)verts2,"surf_simplify");
            }else n1=n;
            first=0;
        } while (n1>5);
    ENDFACES
}

void surf_clip(surf s,double x,double y,double z,
    double nx,double ny,double nz)
{
}

void surf_hollow(surf s,double r)
{
    int i,nv,ne,nf,vi,vi0,vi1,ei,fi;
    surf s1;
    vert v,v0,v1,*verts0,*verts1,*verts2;
    edge e,e0,e1,*edges0,*edges1,*edges2,*edges3,*edges4;
    face f;
    double *nxtab,*nytab,*nztab,*nqtab,nx,ny,nz,nq;

    s1=surf_create();
    surf_size(s,&nv,&ne,&nf);

    // verts
    surf_getnorms(s,&nxtab,&nytab,&nztab,&nqtab);
    verts0=(vert *)mem_get(nv*sizeof(vert),"surf_hollow");
    verts1=(vert *)mem_get(nv*sizeof(vert),"surf_hollow");
    verts2=(vert *)mem_get(ne*sizeof(vert),"surf_hollow");
    vi=0;
    BGNVERTS
        // for modeling
        verts0[vi]=vert_create(s1,v->x,v->y,v->z);
        nq=0.5*nqtab[vi]*r; nx=nxtab[vi]; ny=nytab[vi]; nz=nztab[vi];
        verts1[vi]=vert_create(s1,v->x-nq*nx,v->y-nq*ny,v->z-nq*nz);
        // for mesh-on-surface plots
        //nq=0.25*nqtab[vi]*r; nx=nxtab[vi]; ny=nytab[vi]; nz=nztab[vi];
        //verts0[vi]=vert_create(s1,v->x+nq*nx,v->y+nq*ny,v->z+nq*nz);
        //verts1[vi]=vert_create(s1,v->x-nq*nx,v->y-nq*ny,v->z-nq*nz);
        vi++;
    ENDVERTS
    surf_putnorms(nxtab,nytab,nztab,nqtab);
    nxtab=(double *)mem_get(nf*sizeof(double),"surf_hollow");
    nytab=(double *)mem_get(nf*sizeof(double),"surf_hollow");
    nztab=(double *)mem_get(nf*sizeof(double),"surf_hollow");
    fi=0;
    BGNFACES
        nx=ny=nz=nq=0.0;
        for (e=e0=f->e0;e;e=e1) {
            e1=e->e->ve0;
            v=e->v; nx+=v->x; ny+=v->y; nz+=v->z; nq+=1.0;
            if (e1==e0) e1=(edge)0;
        }
        nxtab[fi]=nx/nq; nytab[fi]=ny/nq; nztab[fi]=nz/nq;
        fi++;
    ENDFACES
    ei=0; nq=1.0-r;
    BGNEDGES
        v=e->v;
        if (e->f) {
            fi=surf_ftoi(e->f);
            nx=nq*v->x+r*nxtab[fi];
            ny=nq*v->y+r*nytab[fi];
            nz=nq*v->z+r*nztab[fi];
        }else{
            vi=surf_vtoi(v); v0=verts0[vi]; v1=verts1[vi];
            nx=(v0->x+v1->x)/2.0; ny=(v0->y+v1->y)/2.0; nz=(v0->z+v1->z)/2.0;
        }
        verts2[ei++]=vert_create(s1,nx,ny,nz);
    ENDEDGES
    mem_put((char *)nxtab,"surf_hollow");
    mem_put((char *)nytab,"surf_hollow");
    mem_put((char *)nztab,"surf_hollow");

    // edges
    edges0=(edge *)mem_get(ne*sizeof(edge),"surf_hollow");
    edges1=(edge *)mem_get(ne*sizeof(edge),"surf_hollow");
    edges2=(edge *)mem_get(ne*sizeof(edge),"surf_hollow");
    edges3=(edge *)mem_get(ne*sizeof(edge),"surf_hollow");
    edges4=(edge *)mem_get(ne*sizeof(edge),"surf_hollow");
    ei=0;
    BGNEDGES
        vi0=surf_vtoi(e->v); vi1=surf_vtoi(e->e->v);
        if (vi0<vi1) {
            e0=edge_create(verts0[vi0],verts0[vi1]);
            e1=edge_create(verts1[vi0],verts1[vi1]);
        }else{ e0=e1=(edge)0; }
        edges0[ei]=e0; edges1[ei]=e1;
        edges2[ei]=edge_create(verts0[vi0],verts2[ei]);
        edges3[ei]=edge_create(verts1[vi0],verts2[ei]);
        edges4[ei]=edge_create(verts2[ei],verts2[surf_etoi(e->e->ve0)]);
        ei++;
    ENDEDGES
    ei=0;
    BGNEDGES
        if (!edges0[ei]) {
            i=surf_etoi(e->e);
            edges0[ei]=edges0[i]->e; edges1[ei]=edges1[i]->e;
        }
        ei++;
    ENDEDGES

    // add edges to verts
    vi=0;
    BGNVERTS
        v0=verts0[vi];
        for (e=e0=v->e0;e;e=e1) {
            e1=e->ve1; ei=surf_etoi(e);
            vert_addedge(v0,edges0[ei]); vert_addedge(v0,edges2[ei]);
            if (e1==e0) e1=(edge)0;
        }
        v1=verts1[vi];
        for (e=e0=v->e0;e;e=e1) {
            e1=e->ve0; ei=surf_etoi(e);
            vert_addedge(v1,edges3[ei]); vert_addedge(v1,edges1[ei]);
            if (e1==e0) e1=(edge)0;
        }
        vi++;
    ENDVERTS
    ei=0;
    BGNEDGES
        v=verts2[ei]; i=surf_etoi(e->ve1->e);
        vert_addedge(v,edges2[ei]); vert_addedge(v,edges4[ei]);
        vert_addedge(v,edges3[ei]); vert_addedge(v,edges4[i]);
        ei++;
    ENDEDGES

    // faces
    surf_createfaces(s1);

    mem_put((char *)verts0,"surf_hollow");
    mem_put((char *)verts1,"surf_hollow");
    mem_put((char *)verts2,"surf_hollow");
    mem_put((char *)edges0,"surf_hollow");
    mem_put((char *)edges1,"surf_hollow");
    mem_put((char *)edges2,"surf_hollow");
    mem_put((char *)edges3,"surf_hollow");
    mem_put((char *)edges4,"surf_hollow");
    surf_swapbody(s,s1);
    surf_destroy(s1);
}

void surf_dual(surf s)
{
    int vi,vi0,vi1,nv,ei,ne,nf;
    vert v,v0,v1,*verts;
    edge e,e0,e1,ex,*edges;
    face f;
    surf s1;
    double x,y,z,n;

    s1=surf_create();

    // vert at each face centroid
    surf_size(s,&nf,&ne,&nv); // NOTE: nf/nv reversed here deliberately!
    verts=(vert *)mem_get(nv*sizeof(vert),"surf_dual");
    vi=0;
    BGNFACES
        x=y=z=n=0.0;
        for (e=e0=f->e0;e;e=e1) {
            e1=e->e->ve0;
            v=e->v; x+=v->x; y+=v->y; z+=v->z; n+=1.0;
            if (e1==e0) e1=(edge)0;
        }
        verts[vi++]=vert_create(s1,x/n,y/n,z/n);
    ENDFACES

    // edge per edge
    edges=(edge *)mem_get(ne*sizeof(edge),"surf_dual");
    ei=0;
    BGNEDGES
        if (e->f && e->e->f) {
            vi0=surf_ftoi(e->f); vi1=surf_ftoi(e->e->f);
            if (vi0<vi1) e1=edge_create(verts[vi0],verts[vi1]);
            else         e1=(edge)0;
        }else e1=(edge)0;
        edges[ei++]=e1;
    ENDEDGES

    // add edges to verts
    vi=0;
    BGNFACES
if (vi!=surf_ftoi(f)) { fprintf(stderr,"vi!=ftoi\n"); exit(1); }
        v=verts[vi++];
        for (e=e0=f->e0;e;e=e1) {
            e1=e->e->ve0;
            ei=surf_etoi(e);
            if (!edges[ei]) ei=surf_etoi(e->e);
            if (ex=edges[ei]) {
                if (ex->v!=v && ex->e->v!=v)
                    { fprintf(stderr,"surf_dual: v not in ex\n"); exit(1); }
                vert_addedge(v,ex);
            }
            if (e1==e0) e1=(edge)0;
        }
    ENDFACES

    // face per vert
    BGNVERTS
        ei=1;
        for (e=e0=v->e0;e;e=e1) {
            e1=e->ve1;
            if (!e->f) ei=0;
            if (e1==e0) e1=(edge)0;
        }
        if (ei) {
            vi0=surf_ftoi(v->e0->f); vi1=surf_ftoi(v->e0->ve1->f);
            v0=verts[vi0]; v1=verts[vi1];
            for (e=e0=v0->e0;e;e=e1) {
                e1=e->ve1;
                if (e->e->v==v1) face_create(e);
                if (e1==e0) e1=(edge)0;
            }
        }
    ENDVERTS

    mem_put((char *)verts,"surf_dual");
    mem_put((char *)edges,"surf_dual");
    surf_swapbody(s,s1);
    surf_destroy(s1);
}

void surf_bound(surf s,double *rxmin,double *rxmax,
    double *rymin,double *rymax,double *rzmin,double *rzmax)
{
    vert v;
    double x,y,z,x0,y0,z0,x1,y1,z1;

    x0=y0=z0=1e24; x1=y1=z1= -1e24;
    BGNVERTS
        x=v->x; y=v->y; z=v->z;
        if (x<x0) x0=x; if (x>x1) x1=x;
        if (y<y0) y0=y; if (y>y1) y1=y;
        if (z<z0) z0=z; if (z>z1) z1=z;
    ENDVERTS
    *rxmin=x0; *rxmax=x1;
    *rymin=y0; *rymax=y1;
    *rzmin=z0; *rzmax=z1;
}

void surf_translate(surf s,double dx,double dy,double dz)
{
    vert v;

    BGNVERTS
        v->x+=dx; v->y+=dy; v->z+=dz;
    ENDVERTS
}

void surf_scale(surf s,double q)
{
    vert v;

    BGNVERTS
        v->x*=q; v->y*=q; v->z*=q;
    ENDVERTS
}

void surf_sharpverts(surf s)
{
    vert v;

    BGNVERTS
        if (!vert_issharp(v)) vert_sharpen(v);
    ENDVERTS
}

void surf_sharpedges(surf s)
{
    edge e;

    BGNEDGES
        if (!edge_issharp(e)) edge_sharpen(e);
    ENDEDGES
}

void surf_unsharpverts(surf s)
{
    vert v;

    BGNVERTS
        if (vert_issharp(v)) vert_unsharpen(v);
    ENDVERTS
}

void surf_unsharpedges(surf s)
{
    edge e;

    BGNEDGES
        if (edge_issharp(e)) edge_unsharpen(e);
    ENDEDGES
}

void surf_sanitize(surf s)
{
    double x0,y0,z0,x1,y1,z1;

    surf_bound(s,&x0,&x1,&y0,&y1,&z0,&z1);
    surf_translate(s,-(x0+x1)/2.0,-(y0+y1)/2.0,-(z0+z1)/2.0);
    x1-=x0; y1-=y0; z1-=z0;
    if (y1>x1) x1=y1; if (z1>x1) x1=z1;
    if (x1<1e-24) return;
    surf_scale(s,1.7/x1);
}

#define NORM(X,Y,Z,R) {                                                       \
    R=sqrt(X*X+Y*Y+Z*Z);                                                      \
    if (R>1e-24) { X/=R; Y/=R; Z/=R; } else { X=1.0; Y=Z=0.0; }               \
}

void surf_vertnorm(vert v,double *rnx,double *rny,double *rnz,double *rnq)
{
    vert v1;
    edge e,e0,e1;
    double x,y,z,xa,ya,za,xb,yb,zb,nx,ny,nz,nq,n,r;

    nx=ny=nz=nq=n=0.0;
    for (e=e0=v->e0;e;e=e1) {
        e1=e->ve1;
        if (e->f) {
            v1=e ->e->v; xa=v1->x-v->x; ya=v1->y-v->y; za=v1->z-v->z;
            v1=e1->e->v; xb=v1->x-v->x; yb=v1->y-v->y; zb=v1->z-v->z;
            x=ya*zb-za*yb; y=za*xb-xa*zb; z=xa*yb-ya*xb;
            NORM(x,y,z,r) nx+=x; ny+=y; nz+=z;
            n+=1.0; nq+=sqrt(xa*xa+ya*ya+za*za);
        }
        if (e1==e0) e1=(edge)0;
    }
    if (n>0.1) nq/=n; else nq=1.0;
    NORM(nx,ny,nz,r)
    *rnx=nx; *rny=ny; *rnz=nz; *rnq=nq;
}

void surf_getnorms(surf s,double **rnx,double **rny,double **rnz,double **rnq)
{
    int nv,vi,ne,nf;
    vert v;
    double *nx,*ny,*nz,*nq;

    surf_size(s,&nv,&ne,&nf);
    nx=(double *)mem_get(nv*sizeof(double),"surf_getnorms");
    ny=(double *)mem_get(nv*sizeof(double),"surf_getnorms");
    nz=(double *)mem_get(nv*sizeof(double),"surf_getnorms");
    nq=(double *)mem_get(nv*sizeof(double),"surf_getnorms");
    vi=0; BGNVERTS surf_vertnorm(v,nx+vi,ny+vi,nz+vi,nq+vi); vi++; ENDVERTS
    *rnx=nx; *rny=ny; *rnz=nz; *rnq=nq;
}

void surf_putnorms(double *nx,double *ny,double *nz,double *nq)
{
    mem_put((char *)nx,"surf_putnorms");
    mem_put((char *)ny,"surf_putnorms");
    mem_put((char *)nz,"surf_putnorms");
    mem_put((char *)nq,"surf_putnorms");
}

void surf_setedgestates(surf s,str action)
{
    int ei,nv,ne,nf,es,esmax;
    edge e,*edges;

    surf_size(s,&nv,&ne,&nf);
    esmax=0;
    edges=(edge *)mem_get(ne*sizeof(edge),"surf_setedgestates");
    ei=0;
    BGNEDGES
        edges[ei++]=(surf_vtoi(e->v)<surf_vtoi(e->e->v)?e:(edge)0);
        es=edge_state(e); if (es>esmax) esmax=es;
    ENDEDGES
    esmax&=0xfffe;
    if (esmax>0) esmax-=2;
    for (ei=0;ei<ne;ei++) {
        if (e=edges[ei]) {
            es=edge_state(e);
            switch (action[0]) {
            case 'U':
                if ((es&0xfffe)>esmax) {
                    edges[ei]=edge_setstate(e,esmax|(es&1));
                }
                break;
            default:
                break;
            }
        }
    }
    mem_put((char *)edges,"surf_setedgestates");
}


