
lib_public {

    // dio.l                            direct (unbuffered) file input/output
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 01-13-98: wrote
    // 02-07-98: fixed bug when a<b
    // 10-18-99: modified to allow compiles when SGI O_DIRECT flag not around
    // 10-21-99: added to sys.l archive
    // 11-04-99: fixed to avoid tmpfile create to get dio params
    // 12-08-99: changed hardwired dio_blk to 0x10000 to fix riptide change
    // 09-09-00: retry without O_DIRECT if it fails (NFS bug found and
    //           tentative patch given by Randy Frank)

    lib_include "../sys.l";

    lib_struct {
        int size0,size;
        char *buf0,*buf;  // use size bytes in buf
    } diobuf;

    diobuf diobuf_create(int size);
    void diobuf_destroy(diobuf dbuf);
    int dio_filesize(char *name);
    diobuf dio_read(char *name,diobuf dbuf);
    int dio_write(char *name,diobuf dbuf);
    int dio_align(int n);
};


int dio_a,dio_blk,dio_max;


lib_init {
/*********************************************************************
    // NOTE: needed to eliminate the tmp file create to get dio params
    // instead use whatever will work on all systems (test it)
    // here is what I observed:
    // riptide:
    //   dio init: F_DIOINFO results d_mem=80 d_miniosz=1000 d_maxiosz=4000000
    // dual:
    //   dio init: F_DIOINFO results d_mem=80 d_miniosz=1000 d_maxiosz=400000
    // hence the choice below this old commented-out section
#ifdef O_DIRECT
    int fd,r,a,b,m;
    struct dioattr d;
    char name[256];

    // NOTE [01-13-98 MAD]: this may be file-system dependent, and
    // therefore this could be wrong!  Need to scan all file systems
    // to get better assurance of correctness...

    // create temporary file so the direct i/o info can be obtained
    sprintf(name,"tmp-dio-init-%d",getpid());
    unlink(name);
    fd=open(name,O_WRONLY|O_CREAT|O_DIRECT,0644);
    if (fd<0) {
        fprintf(stderr,"dio_init: can't open %s\n",name);
        exit(1);
    }
    r=fcntl(fd,F_DIOINFO,&d);
    close(fd);
    unlink(name);
    a=d.d_mem; b=d.d_miniosz; m=d.d_maxiosz;
    printf("dio init: F_DIOINFO results d_mem=%x d_miniosz=%x d_maxiosz=%x\n",
        a,b,m);
    if (a<b) a=b;
    dio_a=a; dio_blk=b; dio_max=m;
#else
    dio_a=4096; dio_blk=4096; dio_max=0x1000000;
#endif
*********************************************************************/
    dio_a=0x10000; dio_blk=dio_a; dio_max=0x400000;
};

diobuf diobuf_create(int size)
{
    int r;
    diobuf dbuf;

    dbuf=(diobuf)mem_get(sizeof(diobuf_struct),"diobuf_create");
    dbuf->size0=size;
    dbuf->size=size=dio_align(size);
    dbuf->buf0=mem_get(size+dio_a+0x10,"diobuf_create");
    r=(int)(dbuf->buf0)+0x10; r=dio_align(r); dbuf->buf=(char *)r;
    return dbuf;
}

void diobuf_destroy(diobuf dbuf)
{
    mem_put(dbuf->buf0,"diobuf_destroy");
    mem_put((char *)dbuf,"diobuf_destroy");
}

int dio_filesize(char *name)
{
    static struct stat st_store;
    static struct stat *st = &st_store;

    if (stat(name,st)) return -1;
    return st->st_size;
}

diobuf dio_read(char *name,diobuf dbuf)
{
    int size,fd,new_dbuf,r,n,k;
    char *buf;

    size=dio_filesize(name); if (size<0) return (diobuf)0;
    if (size!=dio_align(size)) return (diobuf)0;
    if (!dbuf) { dbuf=diobuf_create(size); new_dbuf=1; }
    else new_dbuf=0;

    #ifdef O_DIRECT
        fd=open(name,O_RDONLY|O_DIRECT);
        // if failed with O_DIRECT, maybe will work without
        if (fd<0) fd=open(name,O_RDONLY);
    #else
        fd=open(name,O_RDONLY);
    #endif
    if (fd<0) {
        if (new_dbuf) diobuf_destroy(dbuf);
        return (diobuf)0;
    }

    for (n=size,buf=dbuf->buf;n>0;) {
        k=(n>dio_max?dio_max:n);
        r=read(fd,buf,k);
        if (r!=k) {
            close(fd);
            if (new_dbuf) diobuf_destroy(dbuf);
            return (diobuf)0;
        }
        buf+=k; n-=k;
    }

    close(fd);
    return dbuf;
}

int dio_write(char *name,diobuf dbuf)
{
    int fd,n,k,r;
    char *buf;

    #ifdef O_DIRECT
        fd=open(name,O_WRONLY|O_CREAT|O_DIRECT,0644);
        // if failed with O_DIRECT, maybe will work without
        if (fd<0) fd=open(name,O_WRONLY|O_CREAT,0644);
    #else
        fd=open(name,O_WRONLY|O_CREAT,0644);
    #endif
    if (fd<0) return -1;

    for (n=dbuf->size,buf=dbuf->buf;n>0;) {
        k=(n>dio_max?dio_max:n);
        r=write(fd,buf,k);
        if (r!=k) {
            fprintf(stderr,"dio_write: write() failed r=%d vs. k=%d\n",r,k);
            perror("dio_write");
            close(fd); unlink(name); return -2;
        }
        buf+=k; n-=k;
    }

    close(fd); return 0;
}

int dio_align(int n)
{
    if (n&(dio_a-1)) n+=dio_a-(n&(dio_a-1));
    return n;
}


