
lib_public {

    // tile.l                                               tiled surfaces
    // by Mark Duchaineau (free but copyrighted, see Lib/COPYING)
    //
    // 11-25-99: wrote
    // 12-15-99: added t->ti/titab/eitab fields for tile2qtile application
    // 07-08-00: added tileroamfix remapping application


    lib_include "Sys/sys.l";


    lib_struct {
        int n;                       // number of tiles in object
        tile tiles,tiles1;           // list of tiles (start and end elements)
    } tileobj;

    lib_struct {
        tileobj tob;                 // tile object containing this tile
        tile t0,t1;                  // links for tob's tiles list
        int titab[4];                // indices of edge-neighbor tiles
        int eitab[4];                //   back-pointing edge index
        int nu,nv;                   // tile array sizes
        float *p;                    // array of points (stored x y z x y z ..)
        float *pn;                   // array of normals
        int ti;                      // tile index
    } tile;


    tileobj tileobj_create();
    void tileobj_destroy(tileobj tob);
    tileobj tileobj_readascii(char *filename);
    tile tile_create(tileobj tob,int nu,int nv,float *p);
    void tile_destroy(tile t);

};


lib_application "Tilespin/tilespin.l";
//lib_application "Tileroamfix/tileroamfix.l";


tileobj tileobj_create()
{
    tileobj tob;

    tob=(tileobj)mem_get(sizeof(tileobj_struct),"tileobj_create");
    tob->n=0;
    tob->tiles=(tile)0;
    return tob;
}

void tileobj_destroy(tileobj tob)
{
    while (tob->tiles) tile_destroy(tob->tiles);
    mem_put((char *)tob,"tileobj_destroy");
}

#define ERREXIT { tileobj_destroy(tob); fclose(fp); return (tileobj)0; }

tileobj tileobj_readascii(char *filename)
{
    int n,nu,nv,np,pi,titab[4],eitab[4],i,ti;
    float x,y,z,r,g,b,nx,ny,nz,*pp,*ppn;
    tileobj tob;
    tile t;
    char buf[256];
    FILE *fp;

    if (!(fp=fopen(filename,"r"))) return (tileobj)0;
    n= -1; ti=0;
    while (fgets(buf,255,fp)) {
        switch (buf[0]) {
        case 'n':
            if (n>=0) ERREXIT
            sscanf(buf+1,"%d",&n);
            if (n<1 || n>1000000) ERREXIT
            tob=tileobj_create();
            break;
        case 'f':
            if (n<0) ERREXIT
            sscanf(buf+1,"%d %d %d %d %d %d %d %d",
                titab+0,eitab+0, titab+1,eitab+1,
                titab+2,eitab+2, titab+3,eitab+3);
            break;
        case 'a':
            if (n<0) ERREXIT
            nu=nv= -1;
            sscanf(buf+1,"%d %d",&nu,&nv);
            if (nu<1 || nv<1) ERREXIT
            t=tile_create(tob,nu,nv,(float *)0);
            t->ti=ti++;
            for (i=0;i<4;i++) {
                t->titab[i]=titab[i];
                t->eitab[i]=eitab[i];
            }
            np=nu*nv; pi=0; pp=t->p; ppn=t->pn;
            break;
        case 'v':
            if (n<0) ERREXIT
            if (pi>=np) ERREXIT
            sscanf(buf+1,"%f %f %f %f %f %f %f %f %f",
                &x,&y,&z,&r,&g,&b,&nx,&ny,&nz);
            *pp++ =x; *pp++ =y; *pp++ =z;
            *ppn++ =nx; *ppn++ =ny; *ppn++ =nz;
            pi++;
            break;
        default:
            break;
        }
    }
    fclose(fp);
    return tob;
}

tile tile_create(tileobj tob,int nu,int nv,float *p)
{
    int i,n;
    tile t;

    tob->n++;
    t=(tile)mem_get(sizeof(tile_struct),"tile_create");
    t->t0=tob->tiles1; t->t1=(tile)0;
    if (t->t0) t->t0->t1=t; else tob->tiles=t;
    if (t->t1) t->t1->t0=t; else tob->tiles1=t;
    t->nu=nu; t->nv=nv;
    n=3*nu*nv;
    mem_noclear();
    t->p=(float *)mem_get(n*sizeof(float),"tile_create");
    if (p) { for (i=0;i<n;i++) t->p[i]=p[i]; }
    t->pn=(float *)mem_get(n*sizeof(float),"tile_create");
    t->ti=0;
    return t;
}

void tile_destroy(tile t)
{
    t->tob->n--;
    if (t->t0) t->t0->t1=t->t1; else t->tob->tiles=t->t1;
    if (t->t1) t->t1->t0=t->t0; else t->tob->tiles1=t->t0;
    mem_put((char *)t->p,"tile_destroy");
    mem_put((char *)t->pn,"tile_destroy");
    mem_put((char *)t,"tile_destroy");
}


